/*
 * Test dataspace operations.
 *
 * Copyright (C) 2020, 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/err.h>

#include <systypes/fcntl.h>
#include <systypes/format.h>

#define FMToffsetd "%" pFMToffset "d"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <fsclient/file.h>
#include <mem/memory_utils.h>



static void show(file_t *file, offset_t step, offset_t sample)
{
  /* Allocate a buffer for sampling from the file. */

  char buf[sample + 1];

  for (offset_t offset = 0; offset < file_populated_span(file); offset += step)
  {
    offset_t remaining = file_populated_span(file) - offset;
    offset_t sample_remaining = remaining < sample ? remaining : sample;

    printf(FMToffsetd " bytes from %p...\n", sample_remaining, (file->memory + offset));
    memcpy(buf, (file->memory + offset), sample_remaining);
    buf[sample_remaining] = '\0';

    offset_t leading = 0;
    char *outbuf = buf;

    /* Skip leading zero bytes. */

    while ((*outbuf == '\0') && (leading < sample_remaining))
    {
      outbuf++;
      leading++;
    }

    offset_t trailing = 0;
    char *outbuf_end = outbuf + strlen(outbuf), *end = buf + sample_remaining - 1;

    while ((*end == '\0') && (end >= outbuf_end))
    {
      end--;
      trailing++;
    }

    printf("----\n%s\n----\n", outbuf);

    printf("[" FMToffsetd " zero bytes at start]\n", leading);
    printf("[%zd bytes shown]\n", strlen(outbuf));
    printf("[" FMToffsetd " bytes after string]\n", sample_remaining - leading - strlen(outbuf));
    printf("[" FMToffsetd " zero bytes at end]\n", trailing);
  }
}

int main(int argc, char *argv[])
{
  if (argc < 4)
  {
    printf("Need filename, step and sample size.\n");
    return 1;
  }

  /* Obtain filename and access parameters. */

  char *filename = argv[1];
  offset_t step = atoi(argv[2]);
  offset_t sample = atoi(argv[3]);

  /* Obtain access to the filesystem. */

  l4_cap_idx_t server = l4re_env_get_cap("server");

  /* Invoke the open method to receive the file reference. */

  file_t file;
  long err = file_open(&file, filename, O_RDWR, server);

  if (err)
  {
    printf("Could not obtain file: %s\n", l4sys_errtostr(err));
    return 1;
  }

  offset_t original_size = file.size;

  /* A region of the file is mapped. Here, the start and length is not provided
     as page-aligned offsets, but the region is nevertheless not masked, and the
     region limits will be page-aligned in any case. */

  err = file_mmap(&file, 10, original_size - 20, 0, 0, file_region_flags(file.flags));

  if (err)
  {
    printf("Could not map file region: %s\n", l4sys_errtostr(err));
    return 1;
  }

  printf("Region " FMToffsetd "..." FMToffsetd " populated size " FMToffsetd
         " size " FMToffsetd ".\n",
         file.start_pos, file.end_pos, file_populated_span(&file), file.size);
  printf("File contents:\n");

  show(&file, step, sample);

  printf("File shown.\n");
  printf("Expected...\n"
         "[0 zero bytes at start]\n"
         "[" FMToffsetd " bytes shown]\n"
         "[0 bytes after string]\n"
         "[0 zero bytes at end]\n\n",
         original_size);

  /* A region of the file is mapped. Here, the resulting offsets will be used to
     define a masked region. */

  err = file_mmap(&file, 10, original_size - 20, 10, original_size - 10, file_region_flags(file.flags));

  if (err)
  {
    printf("Could not map file region: %s\n", l4sys_errtostr(err));
    return 1;
  }

  printf("Region " FMToffsetd "..." FMToffsetd " populated size " FMToffsetd
         " size " FMToffsetd ".\n",
         file.start_pos, file.end_pos, file_populated_span(&file), file.size);
  printf("File contents:\n");

  show(&file, step, sample);

  printf("File shown.\n");
  printf("Expected...\n"
         "[10 zero bytes at start]\n"
         "[" FMToffsetd " bytes shown]\n"
         "[" FMToffsetd " bytes after string]\n"
         "[" FMToffsetd " zero bytes at end]\n\n",
         original_size - 20, 10ULL, 10ULL);

  /* Resizing must occur before writing beyond the end of file. Otherwise, the
     data may get discarded if the supporting flexpage needs to be flushed. */

  offset_t old_populated_size = file_populated_span(&file);
  offset_t old_size = file.size;
  offset_t old_populated_size_rounded = round(old_populated_size, page(1));
  offset_t new_size = old_populated_size_rounded + old_populated_size;
  offset_t new_region = round(new_size, page(1));

  printf("Resize to " FMToffsetd "...\n", new_size);

  err = file_resize(&file, new_size);

  if (err)
  {
    printf("Could not resize file: %s\n", l4sys_errtostr(err));
    return 1;
  }

  printf("Resized file...\n");
  printf("Region " FMToffsetd "..." FMToffsetd " populated size " FMToffsetd
         " size " FMToffsetd ".\n",
         file.start_pos, file.end_pos, file_populated_span(&file), file.size);
  printf("Expected " FMToffsetd "..." FMToffsetd " populated size " FMToffsetd
         " size " FMToffsetd ".\n\n",
         0ULL, old_populated_size_rounded, old_populated_size_rounded, new_size);

  /* Re-map to avoid masking the region. */

  err = file_mmap(&file, 0, new_size, 0, 0, file_region_flags(file.flags));

  if (err)
  {
    printf("Could not map file region: %s\n", l4sys_errtostr(err));
    return 1;
  }

  printf("Remapped file without masking...\n");
  printf("Region " FMToffsetd "..." FMToffsetd " populated size " FMToffsetd
         " size " FMToffsetd ".\n",
         file.start_pos, file.end_pos, file_populated_span(&file), file.size);
  printf("Expected " FMToffsetd "..." FMToffsetd " populated size " FMToffsetd
         " size " FMToffsetd ".\n\n",
         0ULL, new_region, old_populated_size, old_size);

  printf("Resize to " FMToffsetd "...\n", new_size);

  err = file_resize(&file, new_size);

  if (err)
  {
    printf("Could not resize file: %s\n", l4sys_errtostr(err));
    return 1;
  }

  printf("Resized file...\n");
  printf("Region " FMToffsetd "..." FMToffsetd " populated size " FMToffsetd
         " size " FMToffsetd ".\n",
         file.start_pos, file.end_pos, file_populated_span(&file), file.size);
  printf("Expected " FMToffsetd "..." FMToffsetd " populated size " FMToffsetd
         " size " FMToffsetd ".\n\n",
         0ULL, round(new_size, page(1)), new_size, new_size);

  /* Copy the file data to another file region. */

  printf("Copy data to " FMToffsetd "...\n", old_populated_size_rounded);

  memcpy(file.memory + old_populated_size_rounded, file.memory, old_size);

  printf("Region " FMToffsetd "..." FMToffsetd " populated size " FMToffsetd
         " size " FMToffsetd ".\n",
         file.start_pos, file.end_pos, file_populated_span(&file), file.size);
  printf("File contents:\n");

  show(&file, step, sample);

  printf("File shown.\n");
  printf("Expected...\n"
         "[0 zero bytes at start]\n"
         "[" FMToffsetd " bytes shown]\n"
         "[" FMToffsetd " bytes after string]\n"
         "[" FMToffsetd " zero bytes at end]\n"
         "...\n"
         "[0 zero bytes at start]\n"
         "[" FMToffsetd " bytes shown]\n"
         "[0 bytes after string]\n"
         "[0 zero bytes at end]\n\n",
         original_size, sample - original_size, sample - original_size,
         original_size);

  /* Re-map to mask the region again. */

  err = file_mmap(&file, 0, new_size, 10, new_size - 10, file_region_flags(file.flags));

  if (err)
  {
    printf("Could not map file region: %s\n", l4sys_errtostr(err));
    return 1;
  }

  printf("Region " FMToffsetd "..." FMToffsetd " populated size " FMToffsetd
         " size " FMToffsetd ".\n",
         file.start_pos, file.end_pos, file_populated_span(&file), file.size);
  printf("File contents:\n");

  show(&file, step, sample);

  printf("File shown.\n");
  printf("Expected...\n"
         "[10 zero bytes at start]\n"
         "[" FMToffsetd " bytes shown]\n"
         "[" FMToffsetd " bytes after string]\n"
         "[" FMToffsetd " zero bytes at end]\n"
         "...\n"
         "[0 zero bytes at start]\n"
         "[" FMToffsetd " bytes shown]\n"
         "[10 bytes after string]\n"
         "[10 zero bytes at end]\n",
         original_size - 10, sample - original_size, sample - original_size,
         original_size - 10);

  printf("End of test.\n");
  return 0;
}

// vim: tabstop=2 expandtab shiftwidth=2
